// ObjReader
// GDImage/OpenGL project
// Created by Patrice TERRIER on january 27, 2015
// copyright www.zapsolution.com
// Last revision 03-03-2020

// Version Lite 3.00

#include <SDKDDKVer.h>
#include <stdio.h>
#include <tchar.h>

#include <windows.h>
#include <windowsx.h>
#include <Commctrl.h>
#include <string>
#include <Shlwapi.h>
#include <Dwmapi.h>
#include <map>
#include <sstream>
#include <algorithm>
#include <vector>

#pragma warning(disable: 4996) // remove Unsafe notifications
#pragma warning(disable: 4312) // remove warning C4312: 'type cast': conversion from 'long' to 'HMENU' of greater size
#pragma warning(disable: 4995) // remove warning C4995: #pragma deprecated

#include <uxtheme.h>
#include <strsafe.h>

#include "WinLIFT.h"
#include "GDImage.h"
#include "ORDLL64.h"
#include "globals.h"
#include "constants.h"
#include "Tools.h"

#pragma comment(lib, "GDImage64.lib")
#pragma comment(lib, "WinLIFT64.lib")
#pragma comment(lib, "ORDLL64.lib")

// Enable tooltip
#pragma comment(linker,"\"/manifestdependency:type='win32' \
name='Microsoft.Windows.Common-Controls' version='6.0.0.0' \
processorArchitecture='*' publicKeyToken='6595b64144ccf1df' language='*'\"")


HWND addLabel(IN HWND hParent, IN long nID, IN WCHAR* zCaption, IN long x, IN long y, IN long w, IN long h, IN DWORD dwStyle, IN int anchor) {
    HWND hCtrl = CreateWindowEx(0, L"Static", zCaption, dwStyle, x, y, w, h, hParent, (HMENU)nID, gP.hInstance, NULL);
    if (IsWindow(hCtrl)) {
        SendMessage(hCtrl, WM_SETFONT, (WPARAM)gP.hFont, 0);
        ZI_SetAnchorMode (hCtrl, anchor);
    }
    return hCtrl;
}

HWND addTextBox(IN HWND hParent, IN long nID, IN WCHAR* zCaption, IN long x, IN long y, IN long w, IN long h, IN DWORD dwStyle, IN int anchor) {
    HWND hCtrl = CreateWindowEx(0, L"Edit", zCaption, dwStyle, x, y, w, h, hParent, (HMENU)nID, gP.hInstance, NULL);
    if (IsWindow(hCtrl)) { 
        SendMessage(hCtrl, WM_SETFONT, (WPARAM)gP.hFont, 0);
        ZI_SetAnchorMode (hCtrl, anchor);
    }
    return hCtrl;
}

HWND addCheckBox(IN HWND hParent, IN long nID, IN WCHAR* zCaption, IN long x, IN long y, IN long w, IN long h, IN DWORD dwStyle, IN int anchor) {
    HWND hCtrl = CreateWindowEx(0, L"Button", zCaption, dwStyle, x, y, w, h, hParent, (HMENU)nID, gP.hInstance, NULL);
    if (IsWindow(hCtrl)) {
        SendMessage(hCtrl, WM_SETFONT, (WPARAM)gP.hFont, 0);
        ZI_SetAnchorMode (hCtrl, anchor);
    }
    return hCtrl;
}

HWND addButton(IN HWND hParent, IN long nID, IN WCHAR* zCaption, IN long x, IN long y, IN long w, IN long h, IN DWORD dwStyle, IN int anchor) {
    HWND hCtrl = CreateWindowEx(0, L"Button", zCaption, dwStyle, x, y, w, h, hParent, (HMENU)nID, gP.hInstance, NULL);
    if (IsWindow(hCtrl)) { 
        SendMessage(hCtrl, WM_SETFONT, (WPARAM)gP.hFont, 0);
        ZI_SetAnchorMode (hCtrl, anchor);
    }
    return hCtrl;
}

HWND addImageButton(IN HWND hParent, IN long nID, IN WCHAR* zCaption, IN long x, IN long y, IN long w, IN long h, IN int anchor) {
    HWND hCtrl = CreateWindowEx(0, L"Button", NULL, WS_CHILD | WS_VISIBLE | BS_BITMAP, x, y, w, h, hParent, (HMENU)nID, gP.hInstance, NULL);
    if (IsWindow(hCtrl)) {
        long bmW = 0, bmH = 0;
        HBITMAP hBitmap = ZI_CreateBitmapFromFile(zCaption, bmW, bmH);
        SendMessage(hCtrl, BM_SETIMAGE, IMAGE_BITMAP, (LPARAM) hBitmap);
        ZI_SetAnchorMode (hCtrl, anchor);
        DeleteObject(hBitmap);
    }
    return hCtrl;
}

void OR_ResizeWindow() {
    if (IsWindow(gP.hOR3D)) {
        POINT lp = { 0 }; ClientToScreen(GetDlgItem(gP.hMain, Visual_Template), &lp);
        RECT rw; GetWindowRect(GetDlgItem(gP.hMain, Visual_Template), &rw);
        SetWindowPos(gP.hOR3D, HWND_TOP, rw.left, rw.top, Width(rw), Height(rw), SWP_NOACTIVATE | SWP_SHOWWINDOW);
    }
}

void ClearMarquee() {
    ZD_DeleteObject(IDC_Marquee); ZI_UpdateWindow(gP.hMarquee, 0);
}

void UpdateMarquee() {
    RECT rc = { 0 };
    GetWindowRect(gP.hMarquee, &rc);
    ZD_DrawTextToCtrl(gP.hMarquee, replace(OR_About(), L"\r\n", L" - "), Width(rc), 0, ZD_ARGB(255, 192,192,192), L"ARIAL", 13, IDC_Marquee, ZS_VISIBLE, 0, 0);
    ZI_UpdateWindow(gP.hMarquee, 0);
}

void DrawMarquee(IN DWORD nDelay) {
    static DWORD nCount;
    DWORD nAniTick = GetTickCount();
    if (nCount == 0) nCount = nAniTick;
    if (nAniTick - nCount >= nDelay) { // high time to update!
        long x = 0, y = 0, w = 0, h = 0;
        ZD_GetObjectBound(IDC_Marquee, w, h);
        if (w > 0) {
            RECT rc = { 0 }; GetWindowRect(gP.hMarquee, &rc);

            //BYTE R = rand() % 255;
            //BYTE G = rand() % 255;
            //BYTE B = rand() % 255;
            //ZD_SetObjectARGB(IDC_Marquee, 255, R, G, B);

            ZD_GetObjectXY(IDC_Marquee, x, y);
            x -= 1; if (x < -w) x = Width(rc);
            ZD_SetObjectXY(IDC_Marquee, x, y, TRUE);
        }
        nCount = 0;
    }
}

void OR_Message(IN UINT uMsg, IN WPARAM wParam, IN LPARAM lParam) {
    BOOL bBool = HIWORD(lParam);
    long nItem = LOWORD(lParam);

    switch (wParam) {
    case LOADED:
        gP.bLoaded = TRUE;
        SendMessage(GetDlgItem(gP.hMain, Audio_Volume), TBM_SETPOS, TRUE, OR_GetAudioVolume());
        // Update the marquee with the model info extracted from "credit.txt"
        UpdateMarquee();

        SetFocus(gP.hMain);
        SetWindowText(GetDlgItem(gP.hMain, IDC_VERTICES), OR_Vertices());
        SetWindowText(GetDlgItem(gP.hMain, IDC_TRIANGLES), OR_Triangles());
        SetWindowText(GetDlgItem(gP.hMain, IDC_INDICES), OR_Indices());
        SetWindowText(GetDlgItem(gP.hMain, IDC_MESHES), OR_Meshes());
        SetWindowText(GetDlgItem(gP.hMain, IDC_MATERIALS), OR_Materials());
        SetWindowText(GetDlgItem(gP.hMain, IDC_OBJSIZE), OR_ObjSize());
        SetWindowText(GetDlgItem(gP.hMain, IDC_LOADTIME), OR_LoadTime());
        OR_ResizeWindow();
        break;

    case ENABLED:
        switch (nItem) {
        case ALPHA:
            break;
        case DEMO:
            EnableWindow(GetDlgItem(gP.hMain, INT_Demo), bBool);
            break;
        case PLAYAUDIO:
            EnableWindow(GetDlgItem(gP.hMain, INT_PlayAudio), bBool);
            EnableWindow(GetDlgItem(gP.hMain, Audio_Volume), bBool);
            EnableWindow(GetDlgItem(gP.hMain, INT_SyncAudio), bBool);
            break;
        case SYNCAUDIO:
            EnableWindow(GetDlgItem(gP.hMain, INT_SyncAudio), bBool);
            break;
        case ANIMATION:
            EnableWindow(GetDlgItem(gP.hMain, INT_Animation), bBool);
            break;
        case FPS:
            EnableWindow(GetDlgItem(gP.hMain, INT_FPS), bBool);
            break;
        case TURBO:
            EnableWindow(GetDlgItem(gP.hMain, INT_Turbo), bBool);
            break;
        }
        break;
    case CHECKED:
        switch (nItem) {
        case ALPHA:
            break;
        case DEMO:
            SendMessage(GetDlgItem(gP.hMain, INT_Demo), BM_SETCHECK, bBool ? BST_CHECKED : BST_UNCHECKED, 0);
            break;
        case PLAYAUDIO:
            SendMessage(GetDlgItem(gP.hMain, INT_PlayAudio), BM_SETCHECK, bBool ? BST_CHECKED : BST_UNCHECKED, 0);
            break;
        case SYNCAUDIO:
            SendMessage(GetDlgItem(gP.hMain, INT_SyncAudio), BM_SETCHECK, bBool ? BST_CHECKED : BST_UNCHECKED, 0);
            break;
        case ANIMATION:
            SendMessage(GetDlgItem(gP.hMain, INT_Animation), BM_SETCHECK, bBool ? BST_CHECKED : BST_UNCHECKED, 0);
            break;
        case FPS:
            SendMessage(GetDlgItem(gP.hMain, INT_FPS), BM_SETCHECK, bBool ? BST_CHECKED : BST_UNCHECKED, 0);
            break;
        case TURBO:
            SendMessage(GetDlgItem(gP.hMain, INT_Turbo), BM_SETCHECK, bBool ? BST_CHECKED : BST_UNCHECKED, 0);
            break;
        }
        break;
    case AUDIO_VOL:
        SendMessage(GetDlgItem(gP.hMain, Audio_Volume), TBM_SETPOS, TRUE, lParam);
        break;
    case LOADING:
        SetWindowText(GetDlgItem(gP.hMain, IDC_VERTICES), NULL);
        SetWindowText(GetDlgItem(gP.hMain, IDC_TRIANGLES), NULL);
        SetWindowText(GetDlgItem(gP.hMain, IDC_INDICES), NULL);
        SetWindowText(GetDlgItem(gP.hMain, IDC_MESHES), NULL);
        SetWindowText(GetDlgItem(gP.hMain, IDC_MATERIALS), NULL);
        SetWindowText(GetDlgItem(gP.hMain, IDC_OBJSIZE), NULL);
        SetWindowText(GetDlgItem(gP.hMain, IDC_LOADTIME), NULL);

//        SetWindowText(GetDlgItem(gP.hMain, LIB_About), NULL);
        ClearMarquee();

        WindowRedraw(GetDlgItem(gP.hMain, IDC_VERTICES));

        break;
    case TIMER:
        DrawMarquee(0);
        break;
    }
}

void ProcessCombo() {
     WCHAR zPath[MAX_PATH] = { 0 };
     WCHAR zSearch[MAX_PATH] = { 0 };
     Path_Combine(zPath, EXEpath(), L"3D_models\\"); Add_Str(zPath, Combo_Selected(gP.combo));
     PathAddBackslash(zPath);
     Path_Combine(zSearch, zPath, L"*.obj");
     WIN32_FIND_DATA fd; ClearMemory(&fd, sizeof(fd));
     HANDLE hFind = FindFirstFile(zSearch, &fd);
     if (hFind != INVALID_HANDLE_VALUE) {
         Add_Str(zPath, fd.cFileName);
         OR_ProcessCommandLine(zPath);
     }
}

LRESULT CALLBACK WndProc(IN HWND hWnd, IN UINT uMsg, IN WPARAM wParam, IN LPARAM lParam) {
    long nW = 0, nH = 0, nMeshCount = 0, nValue = 0, nStatus = 0, wmId = 0, wmEvent = 0;
    HWND hCtrl = 0;
    MINMAXINFO* pMM;
    NMLISTVIEW* ptNMLV;
    WCHAR zTxt[MAX_PATH] = { 0 };
    POINT p = { 0 };
    RECT rc = { 0 };
    MONITORINFO tmi = { 0 };
    HMONITOR hMonitor = NULL;
    HDC hDC = 0;

    // OR private Message
    if (uMsg == WM_LOADED) {
        OR_Message(uMsg, wParam, lParam); 
        return 0;
    }

    switch (uMsg) {
    case WM_GETMINMAXINFO:
        pMM = (MINMAXINFO*)lParam;
        pMM->ptMinTrackSize.x = gP.minTrackSizeW;
        pMM->ptMinTrackSize.y = gP.minTrackSizeH;
        break;

    case WM_NOTIFY:
        ptNMLV = (NMLISTVIEW*)lParam;
        hCtrl = ptNMLV->hdr.hwndFrom;
        break;

    case WM_SYSCOMMAND:
        switch (wParam) {
        case SC_MAXIMIZE:
        case SC_RESTORE:
            OR_ResizeWindow();
            break;
        }
        break;
    case WM_MOVE:
    case WM_SIZE:
        OR_ResizeWindow();
        break;

    case WM_COMMAND:
        wmId    = LOWORD(wParam);
        wmEvent = HIWORD(wParam);
        switch (wmId) {

        case ComboProcess:
            ProcessCombo();
            break;
        case ComboModels:
            if (wmEvent == CBN_SELENDOK) {
                // I need to do this to let windows close the combo, before we load the new selected model (critical section)
                PostMessage(hWnd, WM_COMMAND, MAKLNG(ComboProcess, 0), 0);
            }
            break;

        case INT_Demo:
            PostMessage(gP.hOR3D, WM_LOADED, DEMO, Button_GetCheck(GetDlgItem(gP.hMain, INT_Demo)));
            break;

        case INT_PlayAudio:
            PostMessage(gP.hOR3D, WM_LOADED, PLAYAUDIO, Button_GetCheck(GetDlgItem(gP.hMain, INT_PlayAudio)));
            break;

        case INT_SyncAudio:
            PostMessage(gP.hOR3D, WM_LOADED, SYNCAUDIO, Button_GetCheck(GetDlgItem(gP.hMain, INT_SyncAudio)));
            break;

        case INT_Animation:
            PostMessage(gP.hOR3D, WM_LOADED, ANIMATION, Button_GetCheck(GetDlgItem(gP.hMain, INT_Animation)));
            break;

        case INT_FPS:
            PostMessage(gP.hOR3D, WM_LOADED, FPS, Button_GetCheck(GetDlgItem(gP.hMain, INT_FPS)));
            break;

        case INT_Turbo:
            PostMessage(gP.hOR3D, WM_LOADED, TURBO, Button_GetCheck(GetDlgItem(gP.hMain, INT_Turbo)));
            break;

        case BTN_Help:
            ShellExecuteA(gP.hMain, "open", "http://www.objreader.com/index.php?topic=322.msg5970#msg5970", NULL, NULL, SW_SHOW);
            break;

        case BTN_Info:
            if (gP.bLoaded) {
                GetWindowText(gP.hMain, zTxt, strSize(zTxt));
                MessageBox(gP.hMain, OR_About(), zTxt, MB_ICONINFORMATION);
            }
            break;

        case BTN_GPU:
            if (gP.bLoaded) {
                GetWindowText(gP.hMain, zTxt, strSize(zTxt));
                MessageBox(gP.hMain, OR_GPUinfo(), zTxt, MB_ICONINFORMATION);
            }
            break;

        case BTN_FullScreen:
            hMonitor = MonitorFromWindow(gP.hMain, MONITOR_DEFAULTTONEAREST);
            tmi.cbSize = sizeof(tmi);
            GetMonitorInfo(hMonitor, &tmi);
            ShowWindow(gP.hMain, SW_MAXIMIZE);
            MoveWindow(gP.hOR3D, tmi.rcMonitor.left, tmi.rcMonitor.top, abs(Width(tmi.rcMonitor)), abs(Height(tmi.rcMonitor)), 0);
            PutFocusOn(gP.hOR3D);
            break;

        case BTN_Reset:
            OR_ViewReset();
            break;

        default:
            break;
        }
        return 0;

    case WM_HSCROLL:
        hCtrl = (HWND)lParam;
        if (hCtrl == GetDlgItem(hWnd, Audio_Volume)) {
            OR_SetAudioVolume((long) SendMessage(hCtrl, TBM_GETPOS, 0, 0));
        }
        break;

    case WM_DESTROY:
        PostQuitMessage(0);
        return 0;

    default:
        break;
    }
    return DefWindowProc(hWnd, uMsg, wParam, lParam);
}

void DetectModels() {
    if ((gP.list == NULL) || (gP.combo == NULL)) return;

    WIN32_FIND_DATAW fd;

    List_DeleteAll(gP.list);
    SendMessage(gP.combo, CB_RESETCONTENT, 0, 0);

    WCHAR UsePath[MAX_PATH] = { 0 };
    Path_Combine(UsePath, EXEpath(), L"3D_models\\");
    WCHAR sFolder[MAX_PATH] = { 0 };
    Path_Combine(sFolder, UsePath, L"*.*");
    WCHAR UseDir[MAX_PATH] = { 0 };

    long nCount = 0;
    long nItem = 0;
    HANDLE hFind = FindFirstFile(sFolder, &fd);
    if (hFind != INVALID_HANDLE_VALUE) {
        BOOL bFound = TRUE;
        while (bFound) {
            if (CheckStyle(fd.dwFileAttributes, FILE_ATTRIBUTE_DIRECTORY)) {
                if (List_Add(gP.list, fd.cFileName) > -1) ++nCount;
            }
            bFound = FindNextFile(hFind, &fd);
        }
        FindClose(hFind);
    }

    if (nCount > 0) {
        nCount = 0;
        for (long K = 1; K <= List_Count(gP.list); K++) {
            Path_Combine(sFolder, UsePath, List_GetText(gP.list, K)); Add_Str(sFolder, L"\\*.obj");
            hFind = FindFirstFile(sFolder, &fd);
            if (hFind != INVALID_HANDLE_VALUE) {
                wcscopy(UseDir, List_GetText(gP.list, K));
                if (SendMessage(gP.combo, CB_ADDSTRING, 0, (LPARAM) UseDir) > -1) {
                    CharLower(UseDir);
                    nCount += 1;
                    if (_wcsicmp(UseDir, L"welcome") == 0) nItem = nCount;
                }
            }
        }
        if ((nItem == 0) && (nCount > 0)) nItem = 1;
        if (nItem > 0) Combo_SelectPlus(gP.combo, nItem);
    }
}

int WINAPI wWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPTSTR lpCmdLine, int nCmdShow) {

    UNREFERENCED_PARAMETER(hPrevInstance);
    UNREFERENCED_PARAMETER(lpCmdLine);
    MSG msg = { 0 };

    WCHAR zClass[] = L"ORVIEWER64";

    // Detect process without using a Mutex
    HWND hFound = FindWindow(zClass, NULL);
    if (hFound) {
        if (IsIconic(hFound)) { ShowWindow(hFound, SW_RESTORE); }
        SetForegroundWindow(hFound);
        return 0;
    }

    WCHAR szCaption[MAX_PATH];

    long nRet = 0;
    WNDCLASSEX wcx = { 0 };

    gP.hFont = (HFONT)GetStockObject(DEFAULT_GUI_FONT); // Tahoma
    gP.hInstance = hInstance;

    wcx.cbSize = sizeof(wcx);
    long IsInitialized = GetClassInfoEx(gP.hInstance, zClass, &wcx);
    if (!IsInitialized) {
        wcx.style = CS_HREDRAW | CS_VREDRAW;
        wcx.lpfnWndProc = WndProc;
        wcx.cbClsExtra = NULL;
        wcx.cbWndExtra = NULL;
        wcx.hInstance = gP.hInstance;
        wcx.hIcon = LoadIcon(wcx.hInstance, MAKEINTRESOURCE(IDI_ICON1));
        wcx.hCursor = LoadCursor(NULL, IDC_ARROW);
        wcx.hbrBackground = (HBRUSH)COLOR_BTNSHADOW;
        wcx.lpszMenuName = NULL;
        wcx.lpszClassName = zClass;
        wcx.hIconSm = wcx.hIcon;
        if (RegisterClassEx(&wcx)) { IsInitialized = -1; }
    }

    if (IsInitialized) {
        DWORD dwStyle = WS_OVERLAPPEDWINDOW | WS_CLIPSIBLINGS | WS_CLIPCHILDREN;
        DWORD dwExStyle = 0;
        RECT rc = { 0 };
        SetRect(&rc, 0, 0, ClientW, ClientH);
        AdjustWindowRectEx(&rc, dwStyle, FALSE, dwExStyle);

        gP.minTrackSizeW = rc.right - rc.left;
        gP.minTrackSizeH = rc.bottom - rc.top;
        long x = max((GetSystemMetrics(SM_CXSCREEN) - gP.minTrackSizeW) / 2, 0);
        long y = max((GetSystemMetrics(SM_CYSCREEN) - gP.minTrackSizeH) / 2, 0);
        WCHAR zFileImage[MAX_PATH] = { 0 };

        // Create The Window
        wcscopy(szCaption, MAIN_CAPTION); // PAT: 12-13-2018
        gP.hMain = CreateWindowEx(dwExStyle, zClass, szCaption, dwStyle, x, y, gP.minTrackSizeW, gP.minTrackSizeH, 0, 0, gP.hInstance, NULL);
        if (gP.hMain) {
            HWND hCtrl = 0;

            dwStyle = WS_CHILD | WS_VISIBLE | BS_MULTILINE;
            Path_Combine(zFileImage, EXEresource(), L"g_help.png");
            hCtrl = addImageButton(gP.hMain, BTN_Help, zFileImage, 1049, 6, 48, 48, ANCHOR_RIGHT);
            ZI_CreateToolTip(hCtrl, L"  Online help  ");

            Path_Combine(zFileImage, EXEresource(), L"g_info.png");
            hCtrl = addImageButton(gP.hMain, BTN_Info, zFileImage, 1049, 53, 48, 48, ANCHOR_RIGHT);
            ZI_CreateToolTip(hCtrl, L"  Info about the current 3D model  ");

            Path_Combine(zFileImage, EXEresource(), L"g_fullscreen.png");
            hCtrl = addImageButton(gP.hMain, BTN_FullScreen, zFileImage, 1048, 635, 48, 48, ANCHOR_BOTTOM_RIGHT);
            ZI_CreateToolTip(hCtrl, L"  Switch to full screen mode.  \n  Press [Escape] or [Space bar] to restore.  ");

            Path_Combine(zFileImage, EXEresource(), L"g_reset.png");
            hCtrl = addImageButton(gP.hMain, BTN_Reset, zFileImage, 1205, 635, 48, 48, ANCHOR_BOTTOM_RIGHT);
            ZI_CreateToolTip(hCtrl, L"  Reset view  ");

            dwStyle = WS_CHILD | WS_VISIBLE | CBS_AUTOHSCROLL | CBS_DROPDOWNLIST | WS_VSCROLL | CBS_HASSTRINGS;
            gP.combo = CreateWindowEx(0, L"COMBOBOX", NULL, dwStyle, 1049, 140, 203, 24, gP.hMain, (HMENU)ComboModels, gP.hInstance, NULL);
            SendMessage(gP.combo, WM_SETFONT, (WPARAM)gP.hFont, 0);
            ZI_SetAnchorMode(gP.combo, ANCHOR_RIGHT);
            gP.list = CreateWindowEx(0, L"LISTBOX", L"", WS_CHILD, 0, 0, 0, 0, gP.hMain, (HMENU)ListModels, gP.hInstance, NULL);

            dwExStyle = WS_CHILD | WS_VISIBLE | SS_RIGHT;
            dwStyle = WS_CHILD | WS_VISIBLE | ES_LEFT | ES_READONLY | WS_BORDER;
            hCtrl = addLabel(gP.hMain, LAB_VERTICES, L"Vertices", 1095, 178, 60, 13, dwExStyle, ANCHOR_RIGHT);
            hCtrl = addTextBox(gP.hMain, IDC_VERTICES, NULL, 1159, 176, 92, 19, dwStyle, ANCHOR_RIGHT);

            hCtrl = addLabel(gP.hMain, LAB_TRIANGLES, L"Triangles", 1095, 204, 60, 13, dwExStyle, ANCHOR_RIGHT);
            hCtrl = addTextBox(gP.hMain, IDC_TRIANGLES, NULL, 1159, 202, 92, 19, dwStyle, ANCHOR_RIGHT);

            hCtrl = addLabel(gP.hMain, LAB_INDICES, L"Indices", 1095, 230, 60, 13, dwExStyle, ANCHOR_RIGHT);
            hCtrl = addTextBox(gP.hMain, IDC_INDICES, NULL, 1159, 228, 92, 19, dwStyle, ANCHOR_RIGHT);

            hCtrl = addLabel(gP.hMain, LAB_MESHES, L"Meshes", 1095, 256, 60, 13, dwExStyle, ANCHOR_RIGHT);
            hCtrl = addTextBox(gP.hMain, IDC_MESHES, NULL, 1159, 254, 92, 19, dwStyle, ANCHOR_RIGHT);

            hCtrl = addLabel(gP.hMain, LAB_MATERIALS, L"Materials", 1095, 282, 60, 13, dwExStyle, ANCHOR_RIGHT);
            hCtrl = addTextBox(gP.hMain, IDC_MATERIALS, NULL, 1159, 280, 92, 19, dwStyle, ANCHOR_RIGHT);

            hCtrl = addLabel(gP.hMain, LAB_OBJSIZE, L"OBJ size", 1095, 308, 60, 13, dwExStyle, ANCHOR_RIGHT);
            hCtrl = addTextBox(gP.hMain, IDC_OBJSIZE, NULL, 1159, 306, 92, 19, dwStyle, ANCHOR_RIGHT);

            hCtrl = addLabel(gP.hMain, LAB_LOADTIME, L"Loaded in", 1095, 334, 60, 13, dwExStyle, ANCHOR_RIGHT);
            hCtrl = addTextBox(gP.hMain, IDC_LOADTIME, NULL, 1159, 332, 92, 19, dwStyle, ANCHOR_RIGHT);

            dwStyle = WS_CHILD | WS_VISIBLE | BS_AUTOCHECKBOX;
            hCtrl = addCheckBox(gP.hMain, INT_Demo, L"Demo mode", 1062, 447, 86, 24, dwStyle, ANCHOR_BOTTOM_RIGHT);

            hCtrl = addCheckBox(gP.hMain, INT_PlayAudio, L"Play audio", 1062, 477, 86, 24, dwStyle, ANCHOR_BOTTOM_RIGHT);

            hCtrl = addCheckBox(gP.hMain, INT_SyncAudio, L"Sync audio", 1062, 507, 86, 24, dwStyle, ANCHOR_BOTTOM_RIGHT);

            hCtrl = addCheckBox(gP.hMain, INT_Animation, L"Animation", 1062, 537, 86, 24, dwStyle, ANCHOR_BOTTOM_RIGHT);

            hCtrl = addCheckBox(gP.hMain, INT_FPS, L"Show FPS", 1062, 567, 86, 24, dwStyle, ANCHOR_BOTTOM_RIGHT);

            hCtrl = addCheckBox(gP.hMain, INT_Turbo, L"Turbo mode", 1062, 597, 86, 24, dwStyle, ANCHOR_BOTTOM_RIGHT);

            dwStyle = WS_CHILD | WS_VISIBLE | TBS_HORZ | TBS_NOTICKS;
            hCtrl = CreateWindowExW(0, L"msctls_trackbar32", NULL, dwStyle, 1148, 482, 110, 16, gP.hMain, (HMENU)Audio_Volume, gP.hInstance, NULL);
            ZI_SetAnchorMode(hCtrl, ANCHOR_BOTTOM_RIGHT);
            SendMessage(hCtrl, TBM_SETRANGE, TRUE, MAKELONG(0, 100));
            SendMessage(hCtrl, TBM_SETPOS, TRUE, 15);

            //dwStyle = WS_CHILD | WS_VISIBLE | SS_CENTER;
            //hCtrl = addLabel(gP.hMain, LIB_About, L"Info about the current 3D model", 0, 672+2, 1042, 20, dwStyle, ANCHOR_BOTTOM_WIDTH);

            dwStyle = WS_CHILD | WS_VISIBLE;
            hCtrl = CreateWindowEx(0, L"ZIMAGECTRL", NULL, dwStyle, 0, 672, 1042, 20, gP.hMain, (HMENU)IDC_Marquee, gP.hInstance, NULL);
            if (IsWindow(hCtrl)) {
                long UseBackColor = ZD_ColorARGB(255, 0);
                ZI_SetProperty(hCtrl, ZI_GradientTop, UseBackColor);
                ZI_SetProperty(hCtrl, ZI_GradientBottom, UseBackColor);
                ZI_SetAnchorMode(hCtrl, ANCHOR_BOTTOM_WIDTH);
                ZD_DrawTextToCtrl(hCtrl, L"Info about the current 3D model", 0, 1, ZD_ARGB(255, 192,192,192), L"ARIAL", 13, IDC_Marquee, ZS_VISIBLE, 0, 0);
                gP.hMarquee = hCtrl;
            }

            // We use this static template as a guidance for the size of the OR WS_EX_TOOLWINDOW popup
            dwStyle = WS_CHILD | WS_VISIBLE | SS_BLACKRECT;
            hCtrl = CreateWindowEx(0, L"Static", NULL, dwStyle, 0, 0, 1042, 672, gP.hMain, (HMENU)Visual_Template, gP.hInstance, NULL);
            if (IsWindow(hCtrl)) {
                ZI_SetAnchorMode (hCtrl, ANCHOR_HEIGHT_WIDTH);
            }
            POINT lp = { 0 }; ClientToScreen(hCtrl, &lp);
            GetWindowRect(hCtrl, &rc);
            gP.hOR3D = OR_CreateWindow(WS_EX_TOOLWINDOW | WS_EX_ACCEPTFILES, gP.hMain, WS_POPUP | WS_VISIBLE, lp.x, lp.y, Width(rc), Height(rc), 0);
            if (IsWindow(gP.hOR3D)) {
                // Must be done after the OR window has been created
                WCHAR sFile[16] = { 0 };
                switch (OR_DetectGPU()) {
                case NVIDIA:
                   wcscopy(sFile, L"nVidia.png"); break;
                case ATI:
                    wcscopy(sFile, L"ati.png"); break;
                case INTEL:
                   wcscopy(sFile, L"intel.png"); break;
                default:
                    wcscopy(sFile, L"unknown.png"); break;
                }
                Path_Combine(zFileImage, EXEresource(), sFile);
                hCtrl = addImageButton(gP.hMain, BTN_GPU, zFileImage, 1219, 14, 34, 32, ANCHOR_RIGHT);
                ZI_CreateToolTip(hCtrl, L"  GPU information  ");

                DetectModels();
                OR_TimerEnable(TRUE); // Use the OR syncTimer to move the marquee text
            }

            // Show the main window
            ShowWindow(gP.hMain, nCmdShow);
            WindowRedraw(gP.hMain);

            // Load the default Welcome project
            ProcessCombo();

            SetForegroundWindow(gP.hMain);  // Slightly Higher Priority
            SetFocus(gP.hMain);             // Sets Keyboard Focus To The Window

            // Disable the taskbar icon flashing
            FLASHWINFO fi = { 0 };
            fi.cbSize = sizeof(fi);
            fi.hwnd = gP.hMain;
            FlashWindowEx(&fi);

            while (GetMessage(&msg, NULL, 0, 0)) {
                TranslateMessage(&msg);
                DispatchMessage(&msg);
            }

            nRet = (long)msg.wParam;
        }

        UnregisterClass(wcx.lpszClassName, gP.hInstance);
    }
    return nRet;
}
